
#include <stdio.h>
#include <stdarg.h>

// we separate the hyperbolic and the SOM stuff in two base classes:

class h2_node{
   // the base class to make a H2-lattice-node:
   public:
   float fZ1,fZ2;     // node location in projected H2 (Poincare-Model)
   int iIndex;        // this node's serial index
   int iVisited;      // flag
   int iRing;         // nr of ring
   int iRingPos;      // serial position on ring circumference
   int iNumNb;        // nr of lattice neigbors
   h2_node **ppNb;    // pointers to neighbors 
   h2_node(int iIdx, int iOrder) {
      ppNb = CALLOC(iOrder,h2_node*);
      iNumNb = 0;
      iIndex = iIdx; 
   }
   ~h2_node(void) { free(ppNb); }
};

class som_node {
   public:
   // these elements are SOM-related:
   float *pfVec;      // input/output weight vector
   float fSqrDist;    // squared distance to input 
   float fSig;        // mean squared distance to input 
   float fDistPenalty; // local distance penalty to enforce "conscience" 
   float fActivation; // node "activation" (in range [0,1]) 
   float fEpsInp;     // node local input eps 
   float fEpsOut;     // node local output eps 
   float fSigma;      // node local neighborhood sigma
   char *pcLabel;     // node label
   som_node(int iInp, int iOut) {
      pfVec  = CALLOC(iInp+iOut,float);
      for (int i=0; i<iInp+iOut; i++) pfVec[i]=0.;
      fSig = fEpsInp = fEpsOut = 1;
      pcLabel = NULL;
      fDistPenalty=0;
   }
   ~som_node(void) { free(pfVec); }
};

// combine the two to make a h2_node:

class hsom_node: public som_node, public h2_node {
   public:
   hsom_node(int iIdx, int iInp, int iOut, int iOrder) :
   som_node(iInp, iOut), h2_node(iIdx, iOrder) {
   }
};

// a hyperbolic som ("hsom") consists of hsom_nodes plus
// some additional data + methods:

class hsom{

   friend hsom_node;
protected:
   int iMaxRadius;
   int iNodes;   
   int iDimInp;
   int iDimOut;
   int iNumData;
   float **ppfData;  // iNumData ptrs expected
   int iOrder;
   hsom_node **ppNode;
private:
   int iBestNode, iBestNodeOnly;
   float fH2InvEdgeLength;
   float fEpsAuto;
   /* global learning rates: */
   float fGlobalEpsInp, fGlobalEpsOut;
   float fEpsInpDecay, fEpsOutDecay;  // decay rate for fGlobalEpsInp/Out  
   /* global neighborhood range: */
   float fGlobalSigma;
   /* used = product of global & local */
   float fSigmaDecay;  // decay rate for fSig estimates */ 
   float fBeta;      /*  determines mode:
                      fBeta = 0  winner takes all
                      fBeta > 0  exponential weighting:
                                 G_i = f_i/sum_j f_j  with
                                 f_i = exp(-beta*squaredist_i)
                      fBeta < 0  polynomial weighting:
                                 f_i = (squaredist)^beta
                     */
   float *pfG;       // weighting coefficients G_i 
   float fGMin;      // nodes with smaller G values are ignored 
   float *pfNetOut;  // internal storage of last net output
   float *pfDV;
   float *pfMetric;
   float fC1,fC2,fD1,fD2;  // aux params for Disk automorphism
   int legal_reference(int);
//   void adapt_neighborhood(float*,float*,h2_node*,int,float);
public:
   hsom(int,int,int,int);
   virtual ~hsom(void);
   virtual eprintf(FILE*,const char*,...);
   int exec(float*, float*, float*);  // finds winner
   void adapt_nodes(float*,float*);
   float get_eps_inp(void) { return fGlobalEpsInp; }; 
   float get_eps_out(void) { return fGlobalEpsOut; }; 
   float get_sigma(void)   { return fGlobalSigma;  };
   void   set_node_inp(int,float*);
   float *get_node_inp(int);
   void   set_node_out(int,float*);
   float *get_node_out(int);
   void   set_label(int,char*);
   char  *get_label(int);
   int    get_index(float,float);
   void   get_xy(int,float*,float*);
   int  get_num_nodes(void);
   float *get_weights(void);
   void set_eps_inp(float); 
   void set_eps_out(float); 
   void set_eps_inp_decay(float); 
   void set_eps_out_decay(float); 
   void set_eps_auto(float);
   void set_metric(float*); 
   void get_metric(float*); 
   void set_sigma(float); 
   void set_sigma_decay(float); 
   void set_beta(float);
//
   void init_nodes(float,float,float,float);
   void init_sig(float);
   /* methods accessing dataset instances: */
   void set_training_data(int,float**);
   void use_training_data(void);
   void use_test_data(float*);
   void def_remap(float,float,float,float);
   void do_remap(void);
   void *for_nodes(void*,int*);
   void *for_edges(void*,int,int,int*,int*);
};

