/**********************************************************************

  Example of a class "picture" with an element function that uses 
     instances of another class "point" (implemented in demolib5a).

  Each instance x of class pictures has two member functions:

     x.add(class point ...)   to add zero or more points to
                              picture x

     x.list(void)             to list all points in picture x

  This file shows how to implement the class in C and how to wrap
  it for prog_unit use.

***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "demolib5a.h"   /* headers for the point class   */
#include "demolib5b.h"   /* headers for the picture class */

/* NOTE: the C-data type names (e.g. picture_t) and the prog unit
         class names (e.g., "picture") are not forced to be the same:
         these names exist at different levels of the implementation! */
 
/*************************** in demolib.h: *****************************

typedef struct picture_{ 
   int iNumPoints, iMaxPoints; // current and max nr of points  
   point_t **ppPoint;          // pointer array to iNumPoints   
                               // point_t instances             
} picture_t;  // the C-struct to hold an instance of class "picture" 

***********************************************************************/

/****************** the wrapper for the prog_unit: ********************/

   char *NAMESPACE = "pic";
 
   char *picture_c[] = {

      "B_:",                       /* prefix for this file */ 
      "picture(void)",             /* constructor */
      "~picture()",                /* destructor  */
      "void add(class geo.point ...)", 
                                   /* to add points; note that
                                      use of external class needs
                                      name space prefix of package
                                   */
      "void list(void)",           /* list contents */
      "SAVE_picture()",            /* save list contents */
      "LOAD_picture()",            /* load list contents */
      "COPY_picture()",

      /* allow prog_unit check of optional version info: */

      PICTURE_VERSION, POINT_VERSION,

      /* each token holds a string "classname = version". Tokens of
         this structure are "understood" by the prog_unit and will
         be used for version checking (a full consistency check 
         requires one token for the defined class [here: "picture"] 
         and for each class used as an argument type [here: "point"] 
         or as a base class). The "classname=version" token for each 
         class might, e.g., be #defined in its header file.
      */

      NULL };

   char **CLASSES[] = {picture_c, NULL};

#define RESIZE(Data,MaxSize,Type) \
   /* this auxiliary macro changes its arguments!! */ \
   Data = realloc(Data,(MaxSize=2*MaxSize)*sizeof(Type))
 
/*************************************************************************
   class constructor:  
*************************************************************************/

picture_t *B_picture(int *piDim, void **ppvArg) {
   /* arguments unused here, since function is of type (*)(void) */
   picture_t *pData  = (picture_t*) calloc(1,sizeof(picture_t));
   printf("calling constructor picture(void)\n");
   /* set aside initial memory for adding up to 10 points: */
   pData->iMaxPoints = 10;
   pData->ppPoint   = calloc(pData->iMaxPoints,sizeof(point_t));
   return pData;
}

/*************************************************************************
  class destructor:
*************************************************************************/

void B_FREE_picture(picture_t *pData) {
   /* implementation must permit NULL argument! */
   if (pData) {  int k;
      /* free allocated points: */
      for (k=0; k<pData->iNumPoints; k++) free(pData->ppPoint[k]);
      free(pData->ppPoint);
      free(pData);  /* free instance itself */
   }
}

/*************************************************************************
  class save/load functions:  
*************************************************************************/

void B_SAVE_picture(FILE *fp, picture_t *pData) {
   int i;  point_t *pPoint;
   fprintf(fp,"%d\n",pData->iNumPoints);
   for (i=0; i<pData->iNumPoints; i++) {
      pPoint = pData->ppPoint[i]; 
      fprintf(fp,"%g %g\n",pPoint->x,pPoint->y);
   }
}

void B_LOAD_picture(FILE *fp, picture_t *pData) {

   int i, iNumPoints=0;  point_t *pPoint;

   fscanf(fp,"%d",&iNumPoints);

   /* ensure sufficient space: */
   while (iNumPoints >= pData->iMaxPoints) {
      RESIZE(pData->ppPoint, pData->iMaxPoints, point_t);
   }

   /* load the iNumPoints from file: */
   for (i=0; i<iNumPoints; i++) {
      if (i < pData->iNumPoints) {
         pPoint = pData->ppPoint[i];
      } else {
         pPoint = pData->ppPoint[i] = malloc(sizeof(point_t));
      }
      fscanf(fp,"%g %g",&(pPoint->x), &(pPoint->y));
      /* real code would add error checking! */
   }
   /* free any unused points: */
   for (; i<pData->iNumPoints; i++) free(pData->ppPoint[i]);
   pData->iNumPoints = iNumPoints;

}

/*************************************************************************
  class method add():  adds zero or more points to picture
*************************************************************************/

void B_add(int iNumArg, int *piDim, void **ppvArg, picture_t *pData) {

   /* additional argument iNumArg, since variadic function */

   int i,j; point_t *pArg;

   /* just for illustration: */
   printf("called drawing.add(class point ... )\narguments are\n");
   for (i=0; i<iNumArg; i++) {
      pArg = (point_t*) ppvArg[i];
      printf("point #%d:  x=%g y=%g\n",i, pArg->x, pArg->y);
   }

   /* code to add points to picture pData: */

   while (iNumArg + pData->iNumPoints >= pData->iMaxPoints) {
       /* fetch more space: */
       RESIZE(pData->ppPoint,pData->iMaxPoints,float*);
   }
   
   /* finally add the points: */
   for (i=pData->iNumPoints, j=0; i<pData->iNumPoints+iNumArg; i++,j++) {
      pData->ppPoint[i] = memcpy(malloc(sizeof(point_t)),
                                  ppvArg[j],sizeof(point_t));
   } 
   pData->iNumPoints = i;
}

/*************************************************************************
  class method list():  lists all points stored in picture
*************************************************************************/

void B_list(int *piDim, void **ppvArg, picture_t *pData) {
   int i;
   printf("Picture %p has %d points:\n", pData, pData->iNumPoints);
   for (i=0; i<pData->iNumPoints; i++) {
      point_t *pPoint = pData->ppPoint[i];
      printf("point#%d = (%g %g)\n",i,pPoint->x,pPoint->y);
   }
}

/*************************************************************************
  class copy function:  
*************************************************************************/

picture_t *B_COPY_picture(picture_t *pTo, picture_t *pFrom) {
   if (!pTo) pTo = B_picture(NULL,NULL); 
   if (pFrom) {
      /* code to copy points would come here ... */
      fprintf(stderr,
      "Warning from demolib5b: copy operation not fully coded!\n");
   }
   return pTo;
}

