/* @MANPAGE example of template import library "demolib8.so"

TEMPLATE CLASS FOO:

Implements a simple "array store" for an arbitrary class T.

Usage:

     class foo <hello> x(10);

will create an instance x that can store up to 10 instances
of class hello. The following code 

     class hello u(42), v;

     x.set(3,u); 
     u.value = 4711;     
     ...
     v = x.get(3);
     printf("value = %d\n",v.value);

will store an instance u of hello in slot 3 of x, then set the member 
variable value of u to 4711 (which affects the stored instance, since
passing of template class parameters is always by reference), and
finally set v to reference the stored object, checking its value
in the end.
  
*/

#include <malloc.h>
#include <stdio.h>

#define VERBOSE 0   /* set nonzero value for more printout */

#define CALLOC(A,B)  ((B*) calloc((A),sizeof(B)))

void FREE_T(void*);
void *BY_REF_COPY_T(void*,void*);
void *COPY_T(void*);

char *TYPEINFO_T(void);  
/* 

When called within a function that returns a template class type xxxx, 
TYPEINFO_T() will return a ptr to a (private!) string that describes 
the template class in the form

          "template class xxxx <type1,type2,...>"

where type1,2.. are the type names of the actual template parameters.
Each, the string adr as well as the string contents itself are guaranteed 
to be a unique identifier for the particular class/template binding 
combination. Therefore, when an xxxx instance is constructed, it should
call TYPEINFO_T() and store the returned adr for later use in a
COMPATIBLE_xxxx function to prevent the connection of template object
pins that have the same template classes, but different template arg
bindings (which would cause fatal errors). An example is shown below.

NOTE 1: for structs, it may happen that different prog units implement
different definitions for the same struct name typei. If the above
string signature is ambiguous for such a reason, it will become
disambiguated by an appended #number, where number is an extra 
count that then distinguishes among the different template bindings
that could not be distinguished by their name tuple alone.

NOTE 2: the returned string adr points to private memory that
must not be freed, reallocated or otherwise modified!

NOTE 3: when called in places other than a function that returns
a template class type, the return value of TYPEINFO_T() is undefined.

*/ 

char *NAMESPACE_demolib8 = "t";

char *Foo_c[] = {
   "TEMPLATE<T1>",
   "t_:",
   "foo(int dim)",
   "~foo()",
   "readonly int dim",
   "set(int i, T1 x)",
   "COPY_foo()",
   "COMPATIBLE_foo()",
   "DESCRIBE_foo()",
   "T1 get(int i)",
   "void print(void)",
   "void for_all(void (*fun)(int, T1))",
   /* iterates fun() over all "stored" T1-object references, passing
      additionally their int "storage index" as first parameter. */ 
   NULL
};

char *Hello_c[] = {
   "h_:",
   "hello(int dim)",
   "~hello()",
   "COPY_hello()",
   "int value",
   "void print()",
   NULL
};

char *World_c[] = {
   "w_:",
   "world(int dim)",
   "~world()",
   "COPY_world()",
   "int number",
   "void print()",
   NULL
};


char **CLASSES_demolib8[] = { Hello_c, Foo_c, World_c, NULL };

typedef struct foo__t {
   int iDim;
   void **ppEntry;
   char *pcTypeInfo;  /* to hold TYPEINFO_T() id string adr */
} foo_t;

typedef struct hello__t {
   int iValue;
} hello_t;

typedef struct world__t {
   int iNumber;
} world_t;


/******************************************************************

   AUXILIARY STUFF:

   implementation of "normal" classes hello and world to provide
   example type parameters for the template class foo implemented
   further below (only to make example independent on other libs).
   
   In the example, classes hello and world are very uninteresting:
   the differ only by their name; but they may be any two classes. 
   Their implementation is as usual, without any template-related
   things.

******************************************************************/

hello_t *h_hello(int *piDim, int **ppiArg) {
   int iValue = ppiArg[0][0];
   hello_t *pNew = CALLOC(1,hello_t);
   pNew->iValue = iValue;
   return pNew;
}

void h_FREE_hello(hello_t *pA) {
   if (pA) {
      printf("removing instance of \"hello\" with value=%d (adr %p)\n",pA->iValue,pA);
      free(pA);
   }
}

hello_t *h_COPY_hello(hello_t *pTo, hello_t *pFrom) {
   hello_t *pNew = CALLOC(1,hello_t); /* initial instance */
   if (pFrom) {
      memcpy(pNew, pFrom, sizeof(hello_t)); /* copy data */
   }
   if (pTo) {
       /* deliver new object at address pTo: */
       memcpy(pTo, pNew, sizeof(hello_t));
       free(pNew);  /* leaves dup'd contents intact for access through pTo */
   } else pTo = pNew;
   return pTo;
}       


void h_print(int *piDim, void **ppvArg, hello_t *pA) {
   printf("hi! I am an instance of \"hello\" at %p (my value : %d [%p])\n",pA,pA->iValue,
&(pA->iValue));
}

world_t *w_world(int *piDim, int **ppiArg) {
   int iNumber = ppiArg[0][0];
   world_t *pNew = CALLOC(1,world_t);
   pNew->iNumber = iNumber;
   return pNew;
}

void w_FREE_world(world_t *pA) {
   if (pA) {
      if (VERBOSE) printf("removing instance of \"world\" with number=%d (adr %p)\n",pA->iNumber,pA);
      free(pA);
   }
}

world_t *w_COPY_world(world_t *pTo, world_t *pFrom) {
   world_t *pNew = CALLOC(1,world_t); /* initial instance */
   if (pFrom) {
      memcpy(pNew, pFrom, sizeof(world_t)); /* copy data */
   }
   if (pTo) {
       /* deliver new object at address pTo: */
       memcpy(pTo, pNew, sizeof(world_t));
       free(pNew);  /* leaves dup'd contents intact for access through pTo */
   } else pTo = pNew;
   return pTo;
}       

void w_print(int *piDim, void **ppvArg, world_t *pA) {
   printf("hi! I am an instance of \"world\" at %p (my number : %d)\n",pA,pA->iNumber);
}

int *h_value(int *piDim, hello_t *pA) { 
   if (VERBOSE) printf("register hello(=%p).value at %p\n",pA,&(pA->iValue));
   return &(pA->iValue); 
}

int *w_number(int *piDim, world_t *pA) { 
   if (VERBOSE) printf("register world.number at %p\n",&(pA->iNumber));
   return &(pA->iNumber); 
}

/******************************************************************

   IMPLEMENTATION OF TEMPLATE CLASS foo:

   this illustrates the new stuff (mainly usage of how to
   properly use BY_REF_COPY_T and FREE_T) 

******************************************************************/


foo_t *t_foo(int *piDim, int **ppiArg) {
   int iDim = ppiArg[0][0];
   foo_t *pNew = CALLOC(1,foo_t);
   pNew->ppEntry = CALLOC(iDim,void*);
   pNew->iDim = iDim;
   /* fetch unique type id string adr: */
   pNew->pcTypeInfo = TYPEINFO_T();
   /* Note that this is only useful within 
      constructor functions of template classes! */
   return pNew;
}

void t_FREE_foo(foo_t *pA) {
   if (pA) {
      int i;
      /* use FREE_T to free the stored references to the 
         template type objects: (actually, each FREE_T
         just withdraws one reference, only when the last
         reference (globally!) disappears, the object itself
         is freed):
      */
      for (i=0; i<pA->iDim; i++) FREE_T(pA->ppEntry[i]);
      free(pA->ppEntry);
      free(pA);
      /* NO free of pcTypeInfo: is private!! */
   }
}

int t_COMPATIBLE_foo(foo_t *pA, foo_t *pB) {

   /* the minimal compatibility criterion: */

   return pA->pcTypeInfo == pB->pcTypeInfo;

   /* this is sufficient to avoid that instances
      with different type arguments get connected.
      If needed, the above check may become extended 
      by additional, class-specific conditions
   */
}

char *t_DESCRIBE_foo(foo_t *pA) {
   /* pcTypeInfo is also useful as a 
      "minimal class description": */

   return pA->pcTypeInfo;

   /* (again this is available only for template classes) */
}

foo_t *t_this(int *piDim, foo_t *pA) {
   return pA;
}

int *t_dim(int *piDim, foo_t *pA) { return &pA->iDim; }

foo_t *t_COPY_foo(foo_t *pTo, foo_t *pFrom) {
   foo_t *pNew = CALLOC(1,foo_t); /* initial instance */

   if (pFrom) { int i;

      memcpy(pNew, pFrom, sizeof(foo_t)); /* copy data */
      /* this has included copy of pcTypeInfo! */
 
      /* and make independent instances of internal ref data: */
      pNew->ppEntry = CALLOC(pFrom->iDim, void*);

      for (i=0; i<pFrom->iDim; i++) {

         pNew->ppEntry[i] = BY_REF_COPY_T(pNew->ppEntry[i], pFrom->ppEntry[i]);

         /* 
            REMARK1:
 
            pNew->ppEntry[i] = BY_REF_COPY_T(NULL, pFrom->ppEntry[i]);

            would be equivalent, since ppEntry[i] is initialized as NULL 
         
            REMARK2: the above implements a "shallow copy" (similar to the 
            semantics of "clone" in Java. A deep copy (i.e., re-creating an 
            independent set of copies of the referenced objects) would be 
            obtained with:

                FREE_T(pNew->ppEntry[i]);  // withdraw existing reference
                pNew->ppEntry[i] = COPY_T(pFrom->ppEntry[i];  // make & ref copy

            This may cause a serious performance penalty if the objects are 
            complex!
         */ 
      }
   }
   if (pTo) {
       /* deliver new object at address pTo: */
       memcpy(pTo, pNew, sizeof(foo_t));
       free(pNew);  /* leaves dup'd contents intact for access through pTo */
   } else pTo = pNew;
   return pTo;
}
              
void t_print(int *piDim, void **ppvArg, foo_t *pA) {
   int i;
   for (i=0; i<pA->iDim; i++) {
      printf("obj in slot[%d] = %p (%d refs)\n",i,
              ADR_OF_T(pA->ppEntry[i]),REFS_T(pA->ppEntry[i]));
      /* ADR_OF_T() returns adress of referenced object, mainly
         for diagnostic purposes! (Note that ADR_OF_T(p) != p,
         since p *does not point to the object* : instead, 
         p points to an opaque wrapper object that contains
         the reference with further auxiliary information 
         (mainly a ref count) to the true object inside
      */
   }
}

void t_set(int *piDim, void **ppvArg, foo_t *pA) {
   int iIndex = *((int*) ppvArg[0]);
   if (0 <= iIndex && iIndex < pA->iDim) {
      pA->ppEntry[iIndex] = BY_REF_COPY_T(pA->ppEntry[iIndex],ppvArg[1]);  
      /* this is the default usage: establish a reference to each
         passed template (wrapper) parameter.
      */
   }
}
 
void *t_get(int *piDim, void **ppvArg, foo_t *pA) {
   int iIndex = *((int*) ppvArg[0]);
   if (0 <= iIndex && iIndex < pA->iDim) {
      /* return *reference* to the object: */
      return pA->ppEntry[iIndex];  
      /* Note: indeed, pA->ppEntry[iIndex] is not a ptr to the object
         itself, but instead to a wrapper. But the necessary unwrapping
         is done by the caller (inside the prog_unit!). Thus
         DON'T return REF_T(pA->ppEntry[iIndex]) (which *would* point
         to the object!) here!!! */
   } else return NULL;
}

void t_for_all(int *piDim, void **ppvArg, foo_t *pA) { 
   void (*pFunction)(int,void*) = (void(*)(int,void*)) ppvArg[0];
   /* example of an iterator method that iterates function pFunction()
      over all "stored" objects: */
   int i;
   for (i=0; i<pA->iDim; i++) {
      if (VERBOSE) printf("CALLING function for element #%d (%p):\n",i,pA->ppEntry[i]);
      pFunction(i,pA->ppEntry[i]);
   }
   if (VERBOSE) printf("DONE\n");
}


